open Defs

open Unif

open Printer

let parse s = 
  let lexbuf = Lexing.from_string (s^"\n") in
  try
    Parser.term Lexer.token lexbuf
  with
    Parsing.Parse_error -> failwith (Printf.sprintf "Parsing error near: %s (character %i of the input %s)" (Lexing.lexeme lexbuf) (Lexing.lexeme_start lexbuf) s)

let parse_types s = 
  let lexbuf = Lexing.from_string (s^"\n") in
  try
    Parser.types Lexer.token lexbuf
  with
    Parsing.Parse_error -> failwith (Printf.sprintf "Parsing error near: %s (character %i of the input %s)" (Lexing.lexeme lexbuf) (Lexing.lexeme_start lexbuf) s)

let pt = parse_types

let build_request e1 e2 tt =
  let t = pt tt in
  term_to_hnf (parse e1, t),
  term_to_hnf (parse e2, t)

type 'a list_type = List of 'a list_type types | Atom of 'a

    
let depth = ref 10
let nb_nat = ref 1
let nb_nat_case = ref 0
let nb_add = ref 0
let nb_pred = ref 0
let nb_list = ref 0
let unc = ref false
let inter = ref false
let consts = ref []
let rules = ref []
let file = ref ""
		   
let _ = 
  Sys.catch_break true;
  add_type "S" (pt "nat -> nat");
  add_type "0" (pt "nat");
  add_type "Cons" (pt "nat -> nat_list -> nat_list");
  add_type "Nil" (pt "nat_list");
  Arg.parse ["-d", Arg.Int (fun i -> depth := i), 
	     Printf.sprintf "\t set depth (default %i)" !depth;
	     "-n", Arg.Int (fun i -> nb_nat := i), 
	     Printf.sprintf "\t set number of nat_rec used (default %i)" !nb_nat;
	     "-l", Arg.Int (fun i -> nb_list := i), 
	     Printf.sprintf "\t set number of list_rec used (default %i)" !nb_list;
	     "--nb-nat_case", Arg.Int (fun i -> nb_nat_case := i), 
	     Printf.sprintf "\t set number of nat_case used (default %i)" !nb_nat_case;
	     "-a", Arg.Int (fun i -> nb_add := i), 
	     Printf.sprintf "\t set number of add used (default %i)" !nb_add;
	     "-p", Arg.Int (fun i -> nb_pred := i), 
	     Printf.sprintf "\t set number of pred used (default %i)" !nb_pred;
	     "-u", Arg.Set (unc), "\t print uncalculated nodes";
	     "-i", Arg.Set (inter), "\t launch a parser after the search" ] 
    (fun f -> file := f) "Usage:\ncommand [options] [file]\n"

let request1 =
  build_request "?f:nat -> nat -> nat 0 1" "0" "nat"

let request2 =
  build_request "?f:nat -> nat -> nat 3 1" "2" "nat"

let request3 =
  build_request "?f:nat -> nat -> nat 2 2" "0" "nat"

let request4 =
  build_request "?fs:nat -> (nat -> nat) -> nat -> nat 0 (%m.0) 3" "3" "nat"

let request1bis = 
  build_request "!add (?h4:nat->nat->nat->nat->nat (2) (3) (2) (!add (?h4:nat->nat->nat->nat->nat (2) (3) (1) (!add (?h4:nat->nat->nat->nat->nat (2) (3) (0) (?h1:nat->nat->nat (2) (3))) (?h5:nat->nat->nat->nat->nat (2) (3) (0) (?h1:nat->nat->nat (2) (3))))) (?h5:nat->nat->nat->nat->nat (2) (3) (1) (!add (?h4:nat->nat->nat->nat->nat (2) (3) (0) (?h1:nat->nat->nat (2) (3))) (?h5:nat->nat->nat->nat->nat (2) (3) (0) (?h1:nat->nat->nat (2) (3))))))) (?h5:nat->nat->nat->nat->nat (2) (3) (2) (!add (?h4:nat->nat->nat->nat->nat (2) (3) (1) (!add (?h4:nat->nat->nat->nat->nat (2) (3) (0) (?h1:nat->nat->nat (2) (3))) (?h5:nat->nat->nat->nat->nat (2) (3) (0) (?h1:nat->nat->nat (2) (3))))) (?h5:nat->nat->nat->nat->nat (2) (3) (1) (!add (?h4:nat->nat->nat->nat->nat (2) (3) (0) (?h1:nat->nat->nat (2) (3))) (?h5:nat->nat->nat->nat->nat (2) (3) (0) (?h1:nat->nat->nat (2) (3)))))))" "6" "nat"

let request2bis = 
  build_request "!add (?h4 (0) (2) (1) (!add (?h4 (0) (2) (0) (?h1:nat->nat->nat (0) (2))) (?h5:nat->nat->nat->nat->nat (0) (2) (0) (?h1:nat->nat->nat (0) (2))))) (?h5:nat->nat->nat->nat->nat (0) (2) (1) (!add (?h4 (0) (2) (0) (?h1:nat->nat->nat (0) (2))) (?h5:nat->nat->nat->nat->nat (0) (2) (0) (?h1:nat->nat->nat (0) (2)))))" "0" "nat"

let request1ter = 
  build_request "!add (?h4 (2) (3) (2) (!add (?h4 (2) (3) (1) (!add (?h4 (2) (3) (0) (?h1 (2) (3))) (?h1 (2) (3)))) (!add (?h4 (2) (3) (0) (?h1 (2) (3))) (?h1 (2) (3))))) (!add (?h4 (2) (3) (1) (!add (?h4 (2) (3) (0) (?h1 (2) (3))) (?h1 (2) (3)))) (!add (?h4 (2) (3) (0) (?h1 (2) (3))) (?h1 (2) (3))))" "6" "nat"

let request2ter = 
  build_request "!add (?h4 (0) (2) (1) (!add (?h4 (0) (2) (0) (?h1 (0) (2))) (?h1 (0) (2)))) (!add (?h4 (0) (2) (0) (?h1 (0) (2))) (?h1 (0) (2)))" "0" "nat"

let nat_rec_const = 
  "nat_rec", fun a -> [a; Curry (Base "nat", Curry (a, a)); Base "nat"] 
and nat_rec_rules =
  [parse "!nat_rec ?f ?g 0", parse "?f";
   parse "!nat_rec ?f ?g (S ?n)", 
   parse "?g ?n (!nat_rec ?f ?g ?n)"]
let _ = add_constructors "nat_rec" [term_to_hnf (parse "0", pt "nat");
				    term_to_hnf (parse "%x.S x", pt "nat -> nat")];
  add_base_types "nat_rec" (pt "nat")

let nat_case_const = 
  "nat_case", fun a -> [a; Curry (Base "nat", a); Base "nat"] 
and nat_case_rules =
  [parse "!nat_case ?f ?g 0", parse "?f";
   parse "!nat_case ?f ?g (S ?n)", parse "?g ?n"]
let _ = add_constructors "nat_case" [term_to_hnf (parse "0", pt "nat");
				    term_to_hnf (parse "%x.S x", pt "nat -> nat")];
  add_base_types "nat_case" (pt "nat")

let add_const = 
  "add", fun a -> [pt "nat"; pt "nat"]
and add_rules =
  [parse "!add ?n 0", parse "?n";
   parse "!add ?n (S ?m)", parse "!add (S ?n) ?m"
 ]
let _ = add_constructors "add" [term_to_hnf (parse "0", pt "nat");
				term_to_hnf (parse "%x.S x", pt "nat -> nat")];
    add_base_types "add" (pt "nat")
   
let pred_const = 
  "pred", fun a -> [Curry(Base "nat", a)]
and pred_rules =
  [parse "!pred 0", parse "0";
   parse "!pred (S ?m)", parse "?m"
 ]
let _ = add_constructors "pred" [(*term_to_hnf (parse "0", pt "nat");*)
				 (*term_to_hnf (parse "%x.S x", pt "nat -> nat")*)];
    add_base_types "pred" (pt "nat")
   
let list_rec_const = 
  "list_rec", fun a -> [a; 
			Curry (Base "nat", Curry (Base "nat_list", Curry (a, a))); 
			Base "nat_list"] 
and list_rec_rules =
  [parse "!list_rec ?f ?g Nil", parse "?f";
   parse "!list_rec ?f ?g (Cons ?x ?xs)", 
   parse "?g ?x ?xs (!list_rec ?f ?g ?xs)"]
let _ = add_constructors "list_rec" [term_to_hnf (parse "0", pt "nat");
				    term_to_hnf (parse "%x xs.Cons x xs", pt "nat -> nat_list -> nat_list")];
  add_base_types "list_rec" (pt "nat_list")

   

let _ = 
  if !nb_nat > 0 then
    (consts := [nat_rec_const, !nb_nat]; rules := nat_rec_rules);
  if !nb_nat_case > 0 then
    (consts := (nat_case_const, !nb_nat_case):: !consts; rules := nat_case_rules @ !rules);
  if !nb_list > 0 then
    (consts := (list_rec_const, !nb_list):: !consts; rules := list_rec_rules @ !rules);
  if !nb_add > 0 then
    (consts := (add_const, !nb_add):: !consts; rules := add_rules @ !rules);
  if !nb_pred > 0 then
    (consts := (pred_const, !nb_pred):: !consts; rules := pred_rules @ !rules);
  let request_queue = if (!file = "") then (Queue.push (Queue.push (Queue.push Queue.nil request2) request1)request3)
  else
    (
     let q = ref Queue.nil in
     try
       let lexbuf = Lexing.from_channel (open_in !file) in
       while true do
	 let term1 = Parser.term Lexer.token lexbuf in
	 let term2 = Parser.term Lexer.token lexbuf in
	 let types = Parser.types Lexer.token lexbuf in
	 q := Queue.push !q (term_to_hnf (term1, types), 
			     term_to_hnf (term2, types))
       done;
	 !q
     with Lexer.Eof ->
      !q
    )
  in
  let theta = (unif_consts (!consts) (!rules) request_queue 
		 !depth !unc)
  in
  let final_terms = final_subst "f" theta in
  Printf.printf "\027[01;32m---\027[00m\nNodes : Not calculated\nSolutions : %i\n"
    (List.length final_terms);
  flush stdout;
  print_terms (if !inter then parse_print else pretty_print)
    final_terms;
  (*let lexbuf = Lexing.from_channel (open_in "log") in
  pretty_print (Rewriting.beta_nf (Parser.term Lexer.token lexbuf));*)
  print_string "\n\027[01;33m---\027[00m\n";
  flush stdout;
  if !inter then 
    try
      let lexbuf = Lexing.from_channel stdin in
      while true do
	let result = Parser.term Lexer.token lexbuf in
	print_string "\027[01;34m";
	pretty_print result; 
	print_string "\n\027[01;35m";
	pretty_print (Rewriting.rewrite_system_on_left_func Rewriting.beta_hnf
			(!rules) result); 
	print_string "\n\027[0m";
	flush stdout
      done
    with Lexer.Eof ->
      exit 0
 
  
