
type term = 
  | Var of string | App of term * term | Abs of string * term
  | Unk of string * string Defs.types | Rec_const of string


let str_to_int =
  let rec find var i = function
    | [] -> Defs.Const var
    | x::q when x = var -> Defs.Var i
    | _::q -> find var (i+1) q in
  let rec aux context = function 
    | Var s -> find s 0 context
    | App (s, t) -> Defs.App (aux context s, aux context t)
    | Abs (s, t) -> Defs.Abs (aux (s::context) t)
    | Unk (s, t) -> Defs.Unk (s, Defs.rev_uncurryfy t)
    | Rec_const s -> Defs.Rec_const s
  in aux []
    

let nat_from_int =
  let rec aux c = function
    | 0 -> c
    | n -> aux (App (Var "S", c)) (n-1)
  in aux (Var "0")

let rec varlist t = function
  | [] -> t
  | x::q -> varlist (Abs(x, t)) q

let dummy_type = Defs.Base "dummy type"

let type_table : (string, string Defs.types) Hashtbl.t = Hashtbl.create 10

let add_type = Hashtbl.add type_table

let find_type t = 
  try Hashtbl.find type_table t
  with Not_found -> dummy_type
      
