open Defs

(* printing functions *)
let print_types t = 
  let rec print_types = function
    | Curry (a, b) -> 
	print_string "(";
	print_types a;
	print_string ") -> ";
	print_types b
    | Base (_) -> 
	Printf.printf "B"
  in
  print_string ":";
  print_types t

let print_ind i = print_char (char_of_int (i + 96)) 

let print_Ind s = 
  Printf.printf "?%s" s

let pretty_print_type f t = 
  let rec get_nat accu = function
    | Const "0" -> accu
    | App (Const "S", n) -> get_nat (accu+1) n
    | _ -> raise Not_found
  in
  let rec aux i j b = function
    | App (Const "S", n') as n -> (
	try 
	  print_int (get_nat 0 n)
	with Not_found -> print_string "S ("; aux i j b n'; print_string ")" )
    | App (App (Rec_const "nat_rec", f), g) ->
	Printf.printf "f%i: { 0 -> " j;
	aux i (j+1) b f;
	print_string " | S(n) -> ";
	aux i (j+1) b 
	  (Rewriting.beta_nf (App (App (g, Const "n"), 
				   App (Const (Printf.sprintf "f%i" j),
					Const "n"))));
	print_string " }"
    | App (App (Rec_const "add", f), g) -> 
	aux i j b f;
	print_string " + ";
	aux i j b g
    | Const s -> print_string s
    | Rec_const s -> Printf.printf "!%s" s
    | Var j -> print_ind (i - j)
    | App (s, t) -> aux i j b s; if b then 
	(print_string " \027[1m(\027[0m"; aux i j false t; print_string "\027[1m)\027[0m")
else (print_string " ("; aux i j false t; print_string ")")
    | Abs s -> 
	print_string "(%";
	print_ind (i + 1);
	print_string ".";
	aux (i + 1) j b s; 
	print_string ")"
    | Unk (s, (l, t)) -> 
	print_Ind s;
	f (rev_curryfy t l)
  in aux 0 1 true t

let pretty_print t = pretty_print_type (fun _ -> ()) t

let parse_print_type f t = 
  let rec get_nat accu = function
    | Const "0" -> accu
    | App (Const "S", n) -> get_nat (accu+1) n
    | _ -> raise Not_found
  in
  let rec aux i b = function
    | App (Const "S", n') as n -> (
	try 
	  print_int (get_nat 0 n)
	with Not_found -> print_string "S ("; aux i b n'; print_string ")" )
    | Const s -> print_string s
    | Rec_const s -> Printf.printf "!%s" s
    | Var j -> print_ind (i - j)
    | App (s, t) ->  aux i b s; if b then 
	(print_string " \027[1m(\027[0m"; aux i false t; print_string "\027[1m)\027[0m")
    else (print_string " ("; aux i false t; print_string ")")
    | Abs s -> 
	print_string "(%";
	print_ind (i + 1);
	print_string ".";
	aux (i + 1) b s; 
	print_string ")"
    | Unk (s, (l, t)) -> 
	print_Ind s;
	f (rev_curryfy t l)
  in aux 0 true t

let parse_print t = parse_print_type (fun _ -> ()) t
